package mima.domain;

import java.util.ArrayList;
import java.util.Collection;

/**
 * INV: Invariant from super class
 * <br/>
 * INV: <code>getLeftSubNodes() != null</code>
 * <br/>
 * INV: <code>getRightSubNodes() != null</code>
 * <br/>
 * INV: <code>getRightSubNodes().size() >= getLeftSubNodes().size()</code>
 * 
 * @author Virginio Carfagno
 */
public class CentralNode extends Node {

	private static final Frame DEFAULT_FRAME = Frame.CIRCLE;
	private static final String DEFAULT_TITLE = "Edit this node";
	
	
	private Collection _leftSubNodes;
	private Collection _rightSubNodes;
	
	
	/**
	 * POST: <code>getTitle() == DEFAULT_TITLE</code>
	 * <br/>
	 * POST: <code>getFrame() == DEFAULT_FRAME</code>
	 */
	public CentralNode() {
		super(DEFAULT_TITLE, DEFAULT_FRAME);
		
		_leftSubNodes = new ArrayList();
		_rightSubNodes = new ArrayList();
		
		assert getTitle() == DEFAULT_TITLE;
		assert getFrame() == DEFAULT_FRAME;
		invariant();
	}

	
	protected void doAdd(SubNode subNode) {
		checkNotNull(subNode);
		invariant();
		
		if (isAddToRightSubNodes())
			_rightSubNodes.add(subNode);
		else
			_leftSubNodes.add(subNode);
		
		invariant();
	}
	
	protected void doRemove(SubNode subNode) {
		checkNotNull(subNode);
		invariant();
		
		if (_rightSubNodes.contains(subNode))
			_rightSubNodes.remove(subNode);
		else
			_leftSubNodes.remove(subNode);
		
		invariant();
	}

	/**
	 * This method doesn't return <code>_leftSubNodes</code> directly to ensure
	 * the encapsulation and consistency of this object. A manipulation to the returned 
 	 * <code>Collection</code> doesn't have any effects on this object. To add 
 	 * or remove a <code>SubNode</code> to this node you should use the method 
 	 * <code>setSuperNode(Node)</code> or <code>removeFromSuperNode()</code>
 	 * from <code>SubNode</code>.
	 */
	public Collection getLeftSubNodes() {
		// This method doesn't return _leftSubNodes directly to ensure the
		// encapsulation of this class.
		return new ArrayList(_leftSubNodes);
	}

	/**
	 * This method doesn't return <code>_rightSubNodes</code> directly to ensure
	 * the encapsulation and consistency of this object. A manipulation to the returned 
 	 * <code>Collection</code> doesn't have any effects on this object. To add 
 	 * or remove a <code>SubNode</code> to this node you should use the method 
 	 * <code>setSuperNode(Node)</code> or <code>removeFromSuperNode()</code>
 	 * from <code>SubNode</code>.
	 */
	public Collection getRightSubNodes() {
		return new ArrayList(_rightSubNodes);
	}
	
	public Collection getSubNodes() {
		Collection subNodes =  new ArrayList(getLeftSubNodes());
		subNodes.addAll(getRightSubNodes());
		return subNodes;
	}

	protected void invariant() {
		super.invariant();
		assert getLeftSubNodes() != null;
		assert getRightSubNodes() != null;
		assert getRightSubNodes().size() >= getLeftSubNodes().size();
	}

	private boolean isAddToRightSubNodes() {
		return _rightSubNodes.size() <= _leftSubNodes.size();
	}

	protected boolean isRight(SubNode subNode) {
		return _rightSubNodes.contains(subNode);
	}

}
