package mima.domain;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;


/**
 * INV: <code>getFrame() != null</code>
 * <br/>
 * INV: <code>getSubNodesCount() >= 0</code>
 * <br/>
 * INV: <code>getSubNodes() != null</code>
 * <br/>
 * INV: <code>_nodeObservers != null</code>
 * <p/>
 * This class implements the Observer Pattern.
 * 
 * @author Virginio Carfagno
 */
public abstract class Node {

	/**
	 * Enumeration for Node Frame
	 * 
	 * @author Virginio Carfagno
	 */
	public static class Frame {
		
		public static final Frame CIRCLE = new Frame();
		public static final Frame NONE = new Frame();
		public static final Frame RECTANGLE = new Frame();
		public static final Frame UNDERLINE = new Frame();
		
		private Frame() {
		}
		
	}
	
	
	private Frame _frame;
	
	private boolean _isExpanded;
	
	private String _link;
	
	/**
	 * This field is part of the Observer Pattern.
	 */
	private Collection _nodeObservers;
	
	private String _note;
	
	private String _title;
	
	
	/**
	 * PRE: <code>frame != null</code>
	 * <br/>
	 * POST: <code>this.isExpanded() == true</code>
	 * <br/>
	 * POST: <code>this.getFrame() == frame</code>
	 * <br/>
	 * POST: <code>this.getTitle() == title</code>
	 * <br/>
	 * POST: <code>this.getLink() == null</code>
	 * <br/>
	 * POST: <code>this.getNote() == null</code>
	 * 
	 * @throws IllegalArgumentException if <code>frame == null</code>
	 */
	protected Node(String title, Frame frame) {
		checkNotNull(frame);
		_title = title;
		_frame = frame;
		_isExpanded = true;
		_nodeObservers = new HashSet();
	}
	
	
	/**
	 * PRE: <code>subNode != null</code>
	 * <br/>
	 * POST: <code>hasSubNodes() == true</code>
	 * <br/>
	 * POST: <code>hasSubNode(subNode) == true</code>
	 * <br/>
	 * POST: <code>getSubNodesCount() > 0</code>
	 * 
	 * @throws IllegalArgumentException if <code>subNode == null</code>
	 */
	protected void add(SubNode subNode) {
		// add(subNode) implements the Design Pattern Template Method
		
		checkNotNull(subNode);
		invariant();
		
		doAdd(subNode);
		fireNodeChanged();
		
		assert hasSubNodes();
		assert hasSubNode(subNode);
		assert getSubNodesCount() > 0;
		invariant();
	}
	
	private void checkNotNull(Frame frame) {
		if (frame == null)
			throw new IllegalArgumentException("frame can't be null");
	}
	
	private void checkNotNull(NodeObserver observer) {
		if (observer == null)
			throw new IllegalArgumentException("observer can't be null");
	}
	
	protected void checkNotNull(SubNode subNode) {
		if (subNode == null)
			throw new IllegalArgumentException("subNode can't be null");
	}
	
	/**
	 * PRE: <code>observer != null</code>
	 * <br/>
	 * POST: <code>this.isRegisteredObserver(observer) == false</code>
	 * <p/>
	 * This method is part of the Observer Pattern.
	 * 
	 * @throws IllegalArgumentException if <code>observer == null</code>
	 */
	public void deregisterObserver(NodeObserver observer) {
		checkNotNull(observer);
		invariant();
		
		_nodeObservers.remove(observer);
		
		assert !isRegisteredObserver(observer);
		invariant();
	}
	
	protected abstract void doAdd(SubNode subNode);
	
	protected abstract void doRemove(SubNode subNode);
	
	/**
	 * POST: <code>isExpanded() == true</code>
	 */
	public void expand() {
		invariant();
		
		_isExpanded = true;
		fireNodeChanged();
		
		invariant();
	}
	
	/**
	 * POST: <code>isExpanded() == true</code>
	 */
	public void expandAll() {
		// expandAll() implements the Design Pattern Template Method
		
		invariant();
		
		expand();
		Iterator subNodesIterator = getSubNodes().iterator();
		while (subNodesIterator.hasNext()) {
			invariant();
			Node subNode = (Node)subNodesIterator.next();
			subNode.expandAll();
		}
		
		assert _isExpanded;
		invariant();
	}

	/**
	 * This method is part of the Observer Pattern.
	 */
	private void fireNodeChanged() {
		invariant();
		
		Iterator _observerIterator = _nodeObservers.iterator();
		while (_observerIterator.hasNext()) {
			invariant();
			
			NodeObserver observer = (NodeObserver)_observerIterator.next();
			observer.nodeChanged();
		}
		
		invariant();
	}
	
	public Frame getFrame() {
		return _frame;
	}
	
	public String getLink() {
		return _link;
	}
	
	public String getNote() {
		return _note;
	}
	
	/**
	 * POST: Return value != null
	 */
	public abstract Collection getSubNodes();
	
	/**
	 * POST: Return value >= 0
	 */
	public int getSubNodesCount() {
		return getSubNodes().size();
	}
	
	public String getTitle() {
		return _title;
	}
	
	public boolean hasLink() {
		return _link != null && !_link.equals("");
	}
	
	public boolean hasNote() {
		return _note != null && !_note.equals("");
	}
	
	/**
	 * PRE: <code>subNode != null</code>
	 * 
	 * @throws IllegalArgumentException if <code>subNode == null</code>
	 */
	public boolean hasSubNode(SubNode subNode) {
		checkNotNull(subNode);
		return getSubNodes().contains(subNode);
	}
	
	public boolean hasSubNodes() {
		return getSubNodesCount() > 0;
	}
	
	protected void invariant() {
		assert _frame != null;
		assert _nodeObservers != null;
		assert getSubNodesCount() >= 0;
		assert getSubNodes() != null;
	}
	
	public boolean isExpanded() {
		return _isExpanded;
	}

	/**
	 * PRE: <code>observer != null</code>
	 * 
	 * @throws IllegalArgumentException if <code>observer == null</code>
	 */
	public boolean isRegisteredObserver(NodeObserver observer) {
		checkNotNull(observer);
		return _nodeObservers.contains(observer);
	}
	
	protected abstract boolean isRight(SubNode subNode);

	/**
	 * PRE: <code>observer != null</code>
	 * <br/>
	 * POST: <code>isRegisteredObserver(observer) == true</code>
	 * <p/>
	 * This method is part of the Observer Pattern.
	 * 
	 * @throws IllegalArgumentException if <code>observer == null</code>
	 */
	public void registerObserver(NodeObserver observer) {
		checkNotNull(observer);
		
		invariant();
		assert !_nodeObservers.contains(observer);
		
		_nodeObservers.add(observer);
		
		assert _nodeObservers.contains(observer);
		invariant();
	}
	
	/**
	 * PRE: <code>subNode != null</code>
	 * <br/>
	 * POST: <code>hasSubNode(subNode) == false</code>
	 * <br/>
	 * POST: <code>getSubNodesCount()</code> returns one less before method was
	 * invoked 
	 * 
	 * @throws IllegalArgumentException if <code>subNode == null</code>
	 */
	protected void remove(SubNode subNode) {
		// remove(subNode) implements the Design Pattern Template Method
		
		checkNotNull(subNode);
		invariant();
		int subNodesCount = getSubNodesCount();
		
		doRemove(subNode);
		fireNodeChanged();
		
		assert !hasSubNode(subNode);
		assert subNodesCount - getSubNodesCount() == 1;
		invariant();
	}

	/**
	 * PRE: <code>frame != null</code>
	 * <br/>
	 * POST: <code>getFrame() == frame</code>
	 * 
	 * @throws IllegalArgumentException if <code>frame == null</code>
	 */
	public void setFrame(Frame frame) {
		checkNotNull(frame);
		invariant();
		
		_frame = frame;
		fireNodeChanged();
		
		assert getFrame() == frame;
		invariant();
	}
	
	/**
	 * POST: <code>getLink() == link</code>
	 */
	public void setLink(String link) {
		invariant();
		
		_link = link;
		fireNodeChanged();
		
		assert getLink() == link;
		invariant();
	}
	
	/**
	 * POST: <code>getNote() == note</code>
	 */
	public void setNote(String note) {
		invariant();
		
		_note = note;
		fireNodeChanged();
		
		assert getNote() == note;
		invariant();
	}
	
	/**
	 * POST: <code>getTitle() == title</code>
	 */
	public void setTitle(String title) {
		invariant();
		
		_title = title;
		fireNodeChanged();
		
		assert getTitle() == title;
		invariant();
	}
	
	public void toggleExpansion() {
		invariant();
		
		_isExpanded = !_isExpanded; 
		fireNodeChanged();
		
		invariant();
	}

}
