package mima.presentation.gwt.mindmap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import mima.domain.SubNode;
import mima.presentation.gwt.node.NodeView;
import mima.presentation.gwt.node.SubNodeView;

import com.google.gwt.user.client.ui.AbsolutePanel;
import com.google.gwt.user.client.ui.ComplexPanel;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.VerticalPanel;
import com.google.gwt.user.client.ui.Widget;
import com.google.gwt.user.client.ui.HasHorizontalAlignment.HorizontalAlignmentConstant;
import com.gwt.components.client.Canvas;


/**
 * @author Virginio Carfagno
 */
abstract class SubTree extends Tree {

	private static final String VERTICAL_PADDING = "20px";


	static private void addSubTrees(Tree superTree, Collection subNodes, 
			ComplexPanel subTrees, Collection oldSubNodes) {
		Iterator subNodesIterator = subNodes.iterator();
		while (subNodesIterator.hasNext()) {
			SubNode subNode = (SubNode)subNodesIterator.next();
			if (!oldSubNodes.contains(subNode)) {
				SubTree subTree;
				if (subNode.isRight())
					subTree = new RightSubTree(superTree, subNode);
				else
					subTree = new LeftSubTree(superTree, subNode);
				subTrees.add(subTree);
			}
		}
	}

	static private void deregisterObservers(SubTree subTree) {
		subTree.deregisterObserver();
	
		Iterator subTreesIterator = subTree._subTrees.iterator();
		while (subTreesIterator.hasNext()) {
			SubTree subSubTree = (SubTree)subTreesIterator.next();
			deregisterObservers(subSubTree);
		}
	}

	static void redrawAllEdges(AbsolutePanel edges, ComplexPanel subTrees) {
		edges.clear();
		
		Iterator subTreesIterator = subTrees.iterator();
		while (subTreesIterator.hasNext()) {
			SubTree subTree = (SubTree)subTreesIterator.next();
			Widget edge = subTree.createBezierEdge();
			int edgeY = subTree.calculateEdgeY(edges);
			edges.add(edge, 0, edgeY);
			redrawAllEdges(subTree._edges, subTree._subTrees);
		}
	}
	
	static private void removeSubTrees(Collection subNodes, ComplexPanel subTrees, 
			Collection oldSubNodeViews) {
		Iterator oldSubTreesIterator = oldSubNodeViews.iterator();
		while (oldSubTreesIterator.hasNext()) {
			SubTree oldSubTree = (SubTree)oldSubTreesIterator.next();
			if (!subNodes.contains(oldSubTree.getSubNode())) {
				subTrees.remove(oldSubTree);
				deregisterObservers(oldSubTree);	
			}
		}
	}
	
	static void updateTreeStructure(Tree superTree, Collection subNodes, 
			ComplexPanel subTrees) {
		// Sehr ineffizient
		/*subTreesContainer.clear();
		Iterator subNodesIterator = subNodes.iterator();
		while (subNodesIterator.hasNext()) {
			SubNode subNode = (SubNode)subNodesIterator.next();
			SubTree subTree = new SubTree(this, subNode);
			subTreesContainer.add(subTree);
		}*/
		
		// Effizientere Implementierung
		Collection oldSubTrees = new ArrayList();
		Collection oldSubNodes = new ArrayList();
		Iterator subTreesIterator = subTrees.iterator();
		while (subTreesIterator.hasNext()) {
			SubTree subTree = (SubTree)subTreesIterator.next();
			oldSubTrees.add(subTree);
			oldSubNodes.add(subTree.getSubNode());
		}
		
		removeSubTrees(subNodes, subTrees, oldSubTrees);
		addSubTrees(superTree, subNodes, subTrees, oldSubNodes);
	}
	
	
	private AbsolutePanel _edges;
	
	private VerticalPanel _subTrees;
	
	private Tree _superTree;
	
	
	SubTree(Tree superTree, SubNode subNode) {
		super(subNode);
		_superTree = superTree;
		
		_edges = createEdgesContainer();
		_subTrees = createSubTreesContainer();

		composeSubTree();
	}
	
	
	private int calculateEdgeHeight() {
		return _superTree.getAbsoluteEdgeConnectionPointY() - getAbsoluteEdgeConnectionPointY();
	}
	
	private int calculateEdgeY(Widget superTreeEdges) {
		if (calculateIsEdgeUp())
			return getAbsoluteEdgeConnectionPointY() - superTreeEdges.getAbsoluteTop() - 1;
		else
			return _superTree.getAbsoluteEdgeConnectionPointY() - superTreeEdges.getAbsoluteTop();
	}
	
	private boolean calculateIsEdgeUp() {
		return calculateEdgeHeight() >= 0;
	}
	
	protected abstract void composeSubTree();

	private Widget createBezierEdge() {
		int height = Math.abs(calculateEdgeHeight());
		int y1;
		int y2;
		int x1;
		int x2;
		if (calculateIsEdgeUp()) {
			y1 = height;
			y2 = 0;
		} else { // down
			y1 = 0;
			y2 = height;
		}
		
		if (isRightEdge()) {
			x1 = 0;
			x2 = EDGES_WIDTH;
		}
		else { // left
			x1 = EDGES_WIDTH;
			x2 = 0;
		}

		Canvas canvas = new Canvas(EDGES_WIDTH, height);
		canvas.moveTo(x1, y1);
		canvas.bezierCurveTo(x1, y1, x1, y2, x2, y2);
		canvas.setStrokeColor("blue");
		canvas.stroke();
		return canvas;
	}

	protected VerticalPanel createNodeViewContainer() {
		VerticalPanel nodeViewContainer = new VerticalPanel();
		nodeViewContainer.add(getNodeView());
		nodeViewContainer.add(createVerticalPaddingBox());
		return nodeViewContainer;
	}
	
	protected NodeView createRoot() {
		return new SubNodeView(getSubNode());
	}
	
	private VerticalPanel createSubTreesContainer() {
		VerticalPanel subTrees = new VerticalPanel();
		subTrees.setHorizontalAlignment(getSubTreesHorizontalAlignment());
		return subTrees;
	}
	
	private Widget createVerticalPaddingBox() {
		Label verticalPaddingBox = new Label();
		verticalPaddingBox.removeStyleName("gwt-Label");
		verticalPaddingBox.setHeight(VERTICAL_PADDING);
		return verticalPaddingBox;
	}
	
	private void deregisterObserver() {
		getSubNode().deregisterObserver(this);
		getNodeView().deregisterObserver();
	}

	protected void expand(boolean value) {
		_subTrees.setVisible(value);
		_edges.setVisible(value);
	}
	
	protected AbsolutePanel getEdges() {
		return _edges;
	}

	private SubNode getSubNode() {
		return (SubNode)getNode();
	}
	
	protected VerticalPanel getSubTrees() {
		return _subTrees;
	}
	
	protected abstract HorizontalAlignmentConstant getSubTreesHorizontalAlignment();

	protected abstract boolean isRightEdge();
	
	protected void redrawAllEdges() {
		if (_superTree != null)
			_superTree.redrawAllEdges();
	}
	
	protected void updateTreeStructure() {
		updateTreeStructure(this, getSubNode().getSubNodes(), _subTrees);
	}
}
