package mima.presentation.gwt.node;

import com.google.gwt.user.client.Command;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.Composite;
import com.google.gwt.user.client.ui.KeyboardListenerAdapter;
import com.google.gwt.user.client.ui.MenuBar;
import com.google.gwt.user.client.ui.PopupPanel;
import com.google.gwt.user.client.ui.TextBox;
import com.google.gwt.user.client.ui.VerticalPanel;
import com.google.gwt.user.client.ui.Widget;

import mima.domain.Node;
import mima.domain.NodeObserver;
import mima.domain.SubNode;
import mima.presentation.gwt.nodecomponent.NodeComponent;
import mima.presentation.gwt.nodecore.NodeCore;
import mima.presentation.gwt.nodeeditor.NodeEditor;
import mima.presentation.gwt.nodeframes.CircleFrame;
import mima.presentation.gwt.nodeframes.RectangleFrame;
import mima.presentation.gwt.nodeframes.UnderlineFrame;
import mima.presentation.gwt.util.MouseUtilities;
import mima.presentation.gwt.util.WidgetUtilities;


/**
 * @author Virginio Carfagno
 */
public abstract class NodeView extends Composite implements NodeObserver {

	protected abstract static class HidePopupMenuCommand implements Command {
		
		private PopupPanel _popupMenu;
		
		public HidePopupMenuCommand(PopupPanel popupMenu) {
			_popupMenu = popupMenu;
		}
		
		public abstract void doExecute();
		
		public void execute() {
			doExecute();
			_popupMenu.hide();
		}
	}
	

	private static final String COMMAND_TITLE_CREATE_SUBNODE = "Create Subnode";
	private static final String COMMAND_TITLE_EDIT_NODE_TITLE = "Edit Node Title";
	private static final String COMMAND_TITLE_EXPAND_ALL_SUBNODES = "Expand All Subnodes";
	private static final String COMMAND_TITLE_PASTE_NODE = "Paste Node";
	private static final String COMMAND_TITLE_REMOVE_LINK = "Remove Link";
	private static final String COMMAND_TITLE_REMOVE_NOTE = "Remove Note";
	private static final String COMMAND_TITLE_SET_CIRCLE_FRAME = "Circle";
	private static final String COMMAND_TITLE_SET_LINK = "Set Link";
	private static final String COMMAND_TITLE_SET_NOTE = "Set Note";
	private static final String COMMAND_TITLE_SET_RECTANGLE_FRAME = "Rectangle";
	
	protected static SubNode CuttedSubNode;
	
	private static final String FRAME_MENU = "Frame";
	
	protected static boolean IsCreateSubNode;
	
	
	private Node _node;
	private NodeComponent _nodeComponent;
	private VerticalPanel _nodeComponentContainer;
	private NodeCore _nodeCore;

	
	protected NodeView(Node node, boolean isComposeRight) {
		_node = node;
		_node.registerObserver(this);
		_nodeComponentContainer = createNodeComponentContainer();
		_nodeCore = new NodeCore(_node, isComposeRight);
		initWidget(_nodeComponentContainer);
		
		updateNodeFrame();
	}
	
	
	private Command createCreateSubNodeCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				IsCreateSubNode = true;
				new SubNode(_node);
				IsCreateSubNode = false;
				// node will be expanded, so that the user can see the
				// created SubNode 
				_node.expand();
			}
		};
	}
	
	private Command createEditNodeTitleCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				showTitleEditor(false);
			}
		};
	}
	
	private Command createExpandAllSubNodesCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				_node.expandAll();
			}
		};
	}
	
	private MenuBar createFrameMenu(PopupPanel popupMenu) {
		MenuBar frameMenu = new MenuBar(true);
		frameMenu.setAutoOpen(true);
		
		frameMenu.addItem(COMMAND_TITLE_SET_RECTANGLE_FRAME, 
				createSetRectangleFrameCommand(popupMenu));
		frameMenu.addItem(COMMAND_TITLE_SET_CIRCLE_FRAME, 
				createSetCircleFrameCommand(popupMenu));
		
		return frameMenu;
	}
	
	private VerticalPanel createNodeComponentContainer() {
		return new VerticalPanel() {
			{
				sinkEvents(Event.ONCLICK | Event.MOUSEEVENTS);
			}
			
			public void onBrowserEvent(Event event) {
				// Der folgende Event kann nicht in einem ClickListener 
				// verarbeitet werden, weil dort nicht zwischen rechter und
				// linker Maustaste unterschieden werden kann.
				if (MouseUtilities.isOnDown(event) && // isOnClick geht nicht
						MouseUtilities.isRightButton(event)) {
					int x = MouseUtilities.getX(event);
					int y = MouseUtilities.getY(event);
					showPopupMenu(x, y);
				}
				super.onBrowserEvent(event);
			}
		};
	}
	
	private Command createNodeEditorCommand(final PopupPanel popupMenu, final boolean isEditLink) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				new NodeEditor(popupMenu.getPopupLeft(), 
						popupMenu.getPopupTop(), _node, isEditLink);
			}
		};
	}
	
	private NodeComponent createNodeFrame() {
		Node.Frame frame = _node.getFrame();
		if (frame == Node.Frame.CIRCLE)
			return new CircleFrame(_nodeCore);
		else if (frame == Node.Frame.RECTANGLE)
			return new RectangleFrame(_nodeCore);
		else if (frame == Node.Frame.UNDERLINE)
			return new UnderlineFrame(_nodeCore);
		else if (frame == Node.Frame.NONE) 
			return _nodeCore;
		else
			throw new UnsupportedOperationException("Frame type " + frame + " not supported");
	}
	
	private Command createPasteNodeCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				CuttedSubNode.setSuperNode(_node);
				_node.expand();
				CuttedSubNode = null;
			}
		};
	}
	
	private PopupPanel createPopupMenu() {
		final PopupPanel popupPanel = new PopupPanel(true);
		popupPanel.add(createPopupMenu(popupPanel));
		return popupPanel;
	}
	
	protected MenuBar createPopupMenu(PopupPanel popupPanel) {
		MenuBar popupMenu = new MenuBar(true);
		
		popupMenu.addItem(COMMAND_TITLE_CREATE_SUBNODE, createCreateSubNodeCommand(popupPanel));
		
		popupMenu.addItem(COMMAND_TITLE_EDIT_NODE_TITLE, createEditNodeTitleCommand(popupPanel));
		
		if (CuttedSubNode != null)
			popupMenu.addItem(COMMAND_TITLE_PASTE_NODE, createPasteNodeCommand(popupPanel));
		
		popupMenu.addItem(COMMAND_TITLE_SET_LINK, createNodeEditorCommand(popupPanel, true));
		
		if (_node.hasLink())
			popupMenu.addItem(COMMAND_TITLE_REMOVE_LINK, createRemoveLinkCommand(popupPanel));
		
		popupMenu.addItem(COMMAND_TITLE_SET_NOTE, createNodeEditorCommand(popupPanel, false));
		
		if (_node.hasNote())
			popupMenu.addItem(COMMAND_TITLE_REMOVE_NOTE, createRemoveNoteCommand(popupPanel));
		
		popupMenu.addItem(COMMAND_TITLE_EXPAND_ALL_SUBNODES, createExpandAllSubNodesCommand(popupPanel));
			
		popupMenu.addItem(FRAME_MENU, createFrameMenu(popupPanel));
		return popupMenu;
	}
	
	private Command createRemoveLinkCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				_node.setLink(null);
			}
		};
	}
	
	private Command createRemoveNoteCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				_node.setNote(null);
			}
		};
	}
	
	private Command createSetCircleFrameCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				_node.setFrame(Node.Frame.CIRCLE);
			}
		};
	}
	
	private Command createSetRectangleFrameCommand(PopupPanel popupMenu) {
		return new HidePopupMenuCommand(popupMenu) {
			public void doExecute() {
				_node.setFrame(Node.Frame.RECTANGLE);
			}
		};
	}
	
	public void deregisterObserver() {
		_node.deregisterObserver(this);
		_nodeCore.deregisterObserver();
	}
	
	public int getAbsoluteEdgeConnectionPointY() {
		return _nodeComponent.getAbsoluteEdgeConnectionPointY();
	}
	
	protected Node getNode() {
		return _node;
	}

	/**
	 * This method was introduced for testing with Pattern Window Driver.
	 */
	public String getNodeTitle() {
		return _nodeCore.getNodeTitle();
	}
	
	public void nodeChanged() {
		updateNodeFrame();
	}
	
	private void showPopupMenu(int x, int y) {
		PopupPanel popupMenu = createPopupMenu();
		popupMenu.setPopupPosition(x, y);
		popupMenu.show();
	}

	protected void showTitleEditor(final boolean isCreateSubNode) {
		new NodeComponent() {
			
			private NodeComponent _oldNodeComponent;

			{
				TextBox titleEditor = new TextBox() {
					
					{
						addKeyboardListener(new KeyboardListenerAdapter() {
							public void onKeyPress(Widget sender,
									char keyCode, int modifiers) {
								if (keyCode == KEY_ENTER)
									accept();
								else if (keyCode == KEY_ESCAPE)
									cancel();
							}
						});
					}
					
					protected void accept() {
						switchToViewMode();
						_node.setTitle(getText());
					}
					
					protected void cancel() {
						switchToViewMode();
						if (isCreateSubNode)
							((SubNode)_node).removeFromSuperNode();
					}
					
					protected void onLoad() {
						setFocus(true);
					}
					
					private void switchToViewMode() {
						_nodeComponent = _oldNodeComponent;
						updateNodeComponentContainer();
					}
				};
				initWidget(titleEditor);
				
				_oldNodeComponent = _nodeComponent;
				_nodeComponent = this;
				updateNodeComponentContainer();
			}
			
			public int getAbsoluteEdgeConnectionPointY() {
				return WidgetUtilities.getAbsoluteMiddle(this);
			}
			
		};
	}

	private void updateNodeComponentContainer() {
		assert _nodeComponentContainer.getWidgetCount() < 2;
		
		if (_nodeComponentContainer.getWidgetCount() > 0)
			_nodeComponentContainer.remove(0);
		_nodeComponentContainer.add(_nodeComponent);
		
		assert _nodeComponent.getParent() == _nodeComponentContainer;
	}

	private void updateNodeFrame() {
		_nodeComponent = createNodeFrame();
		updateNodeComponentContainer();
	}
}
