package mima.presentation.gwt.nodecore;

import mima.domain.Node;
import mima.domain.NodeObserver;
import mima.presentation.gwt.nodecomponent.NodeComponent;
import mima.presentation.gwt.noteview.NoteView;
import mima.presentation.gwt.util.MouseUtilities;
import mima.presentation.gwt.util.WidgetUtilities;

import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.ClickListener;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Image;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.MouseListener;
import com.google.gwt.user.client.ui.MouseListenerAdapter;
import com.google.gwt.user.client.ui.Widget;


/**
 * @author Virginio Carfagno
 */
public class NodeCore extends NodeComponent implements NodeObserver {

	private static final String COLLAPSED_ICON_URL = "tree_closed.gif";
	private static final String EXPANDED_ICON_URL = "tree_open.gif";
	private static final String LINK_ICON_URL = "link.png";
	private static final String NOTE_ICON_URL = "note.png";
	private static final String STYLE_NAME = "mima-NodeView";
	private static final String STYLE_NAME_SELECTED = "mima-NodeView-selected";
	
	
	private Image _expansionIconButton;
	
	private HorizontalPanel _iconButtons;
	private Widget _linkIconButton;
	
	private Node _node;
	private HorizontalPanel _nodeCore;
	
	private Widget _noteIconButton;
	
	private Label _titleLabel;
	
	
	public NodeCore(Node node, boolean isComposeRight) {
		_node = node;
		_node.registerObserver(this);
		
		_nodeCore = new HorizontalPanel();
		initWidget(_nodeCore);
		
		setStyleName(STYLE_NAME);
		
		_titleLabel = createTitleLabel();
		_iconButtons = new HorizontalPanel();
		
		_expansionIconButton = createExpansionIcon();
		
		composeNodeCore(isComposeRight);
	}
	
	
	private void composeNodeCore(boolean isComposeRight) {
		_nodeCore.add(_titleLabel);
		_nodeCore.add(_iconButtons);
		
		if (isComposeRight)
			positionExpansionIconsToRight();
		else
			positionExpansionIconsToLeft();
	}
	
	private Image createExpansionIcon() {
		Image expansionIcon = createIcon(EXPANDED_ICON_URL);
		expansionIcon.addClickListener(createToggleExpansionListener());
		return expansionIcon;
	}

	private Image createIcon(String iconURL) {
		Image icon = new Image(iconURL);
		icon.addMouseListener(createSelectionListener());
		return icon;
	}
	
	private Widget createLinkIconButton() {
		Image linkIconButton = createIcon(LINK_ICON_URL);
		linkIconButton.addClickListener(new ClickListener() {
			public void onClick(Widget sender) {
				Window.open(_node.getLink(), _node.getTitle(), "");
			}
		});
		return linkIconButton;
	}
	
	private Widget createNoteIconButton() {
		Image noteIconButton = new Image(NOTE_ICON_URL) {
			public void onBrowserEvent(Event event) {
				if (MouseUtilities.isOnClick(event)) {
					new NoteView(_node, MouseUtilities.getX(event), 
							MouseUtilities.getY(event));
				}
				super.onBrowserEvent(event);
			}
		};
		noteIconButton.addMouseListener(createSelectionListener());
		return noteIconButton;
	}

	private MouseListener createSelectionListener() {
		return new MouseListenerAdapter() {
			public void onMouseEnter(Widget titleLabel) {
				titleLabel.addStyleName(STYLE_NAME_SELECTED);
			}
			public void onMouseLeave(Widget titleLabel) {
				titleLabel.removeStyleName(STYLE_NAME_SELECTED);
			}
		};
	}

	private Label createTitleLabel() {
		return new Label(_node.getTitle()) {
			
			{
				addClickListener(createToggleExpansionListener());
				addMouseListener(createSelectionListener());
				setWordWrap(false);
			}
			
			protected void onLoad() {
				fireNodeComponentChanged();
			}
			
		};
	}
	
	private ClickListener createToggleExpansionListener() {
		return new ClickListener() {
			public void onClick(Widget sender) {
				_node.toggleExpansion();
			}
		};
	}

	public void deregisterObserver() {
		_node.deregisterObserver(this);
	}

	public int getAbsoluteEdgeConnectionPointY() {
		return WidgetUtilities.getAbsoluteMiddle(this);
	}

	/**
	 * This method was introduced for testing with Pattern Window Driver.
	 */
	public String getNodeTitle() {
		return _titleLabel.getText();
	}

	public void nodeChanged() {
		updateView();
	}

	protected void onLoad() {
		super.onLoad();
		updateView();
	}
	
	private void positionExpansionIconsToLeft() {
		_nodeCore.insert(_expansionIconButton, 0);
	}

	private void positionExpansionIconsToRight() {
		_nodeCore.add(_expansionIconButton);
	}
	
	private void updateExpansionIcon() {
		String expansionIconUrl;
		if (_node.isExpanded())
			expansionIconUrl = EXPANDED_ICON_URL;
		else
			expansionIconUrl = COLLAPSED_ICON_URL;
		_expansionIconButton.setUrl(expansionIconUrl);
		
		_expansionIconButton.setVisible(_node.hasSubNodes());
	}

	private void updateLinkIconButton() {
		if (_node.hasLink()) {
			if (_linkIconButton == null) {
				_linkIconButton = createLinkIconButton();
				_iconButtons.add(_linkIconButton);
			}
		}
		else {
			if (_linkIconButton != null) {
				_iconButtons.remove(_linkIconButton);
				_linkIconButton = null;
			}
		}
	}
	
	private void updateNoteIconButton() {
		if (_node.hasNote()) {
			if (_noteIconButton == null) {
				_noteIconButton = createNoteIconButton();
				_iconButtons.add(_noteIconButton);
			}
		}
		else {
			if (_noteIconButton != null) {
				_iconButtons.remove(_noteIconButton);
				_noteIconButton = null;
			}
		}
	}
	
	private void updateTitle() {
		_titleLabel.setText(_node.getTitle());
	}

	private void updateView() {
		updateTitle();
		updateLinkIconButton();
		updateNoteIconButton();
		updateExpansionIcon();
		fireNodeComponentChanged();
	}
	
}
